@@ -37,6 +37,7 @@
 #include "kernel-shared/volumes.h"
 #include "kernel-shared/transaction.h"
 #include "kernel-shared/zoned.h"
+#include "crypto/crc32c.h"
 #include "common/utils.h"
 #include "common/path-utils.h"
 #include "common/device-utils.h"
@@ -59,7 +60,7 @@ struct mkfs_allocation {
 	u64 system;
 };
 
-static int create_metadata_block_groups(struct btrfs_root *root, int mixed,
+static int create_metadata_block_groups(struct btrfs_root *root, bool mixed,
 				struct mkfs_allocation *allocation)
 {
 	struct btrfs_fs_info *fs_info = root->fs_info;
@@ -139,7 +140,7 @@ err:
 }
 
 static int create_data_block_groups(struct btrfs_trans_handle *trans,
-		struct btrfs_root *root, int mixed,
+		struct btrfs_root *root, bool mixed,
 		struct mkfs_allocation *allocation)
 {
 	struct btrfs_fs_info *fs_info = root->fs_info;
@@ -208,22 +209,76 @@ err:
 	return ret;
 }
 
-static int __recow_root(struct btrfs_trans_handle *trans,
-			 struct btrfs_root *root)
+static int __recow_root(struct btrfs_trans_handle *trans, struct btrfs_root *root)
 {
-	struct extent_buffer *tmp;
+	struct btrfs_path path;
+	struct btrfs_key key;
 	int ret;
 
-	if (trans->transid != btrfs_root_generation(&root->root_item)) {
-		extent_buffer_get(root->node);
-		ret = __btrfs_cow_block(trans, root, root->node,
-					NULL, 0, &tmp, 0, 0);
+	btrfs_init_path(&path);
+	key.objectid = 0;
+	key.type = 0;
+	key.offset = 0;
+
+	/* Get a path to the left-most leaves */
+	ret = btrfs_search_slot(NULL, root, &key, &path, 0, 0);
+	if (ret < 0)
+		return ret;
+
+	while (true) {
+		struct btrfs_key found_key;
+
+		/*
+		 * Our parent nodes must not be newer than the leaf, thus if
+		 * the leaf is as new as the transaction, no need to re-COW.
+		 */
+		if (btrfs_header_generation(path.nodes[0]) == trans->transid)
+			goto next;
+
+		/*
+		 * Grab the key of current tree block and do a COW search to
+		 * the current tree block.
+		 */
+		btrfs_item_key_to_cpu(path.nodes[0], &key, 0);
+		btrfs_release_path(&path);
+
+		/* This will ensure this leaf and all its parent get COWed */
+		ret = btrfs_search_slot(trans, root, &key, &path, 0, 1);
+		if (ret < 0)
+			goto out;
+		ret = 0;
+		btrfs_item_key_to_cpu(path.nodes[0], &found_key, 0);
+		ASSERT(btrfs_comp_cpu_keys(&key, &found_key) == 0);
+
+next:
+		ret = btrfs_next_leaf(root, &path);
+		if (ret < 0)
+			goto out;
+		if (ret > 0) {
+			ret = 0;
+			goto out;
+		}
+	}
+out:
+	btrfs_release_path(&path);
+	return ret;
+}
+
+static int recow_global_roots(struct btrfs_trans_handle *trans)
+{
+	struct btrfs_fs_info *fs_info = trans->fs_info;
+	struct btrfs_root *root;
+	struct rb_node *n;
+	int ret = 0;
+
+	for (n = rb_first(&fs_info->global_roots_tree); n; n = rb_next(n)) {
+		root = rb_entry(n, struct btrfs_root, rb_node);
+		ret = __recow_root(trans, root);
 		if (ret)
 			return ret;
-		free_extent_buffer(tmp);
 	}
 
-	return 0;
+	return ret;
 }
 
 static int recow_roots(struct btrfs_trans_handle *trans,
@@ -236,9 +291,6 @@ static int recow_roots(struct btrfs_trans_handle *trans,
 	if (ret)
 		return ret;
 	ret = __recow_root(trans, info->tree_root);
-	if (ret)
-		return ret;
-	ret = __recow_root(trans, info->extent_root);
 	if (ret)
 		return ret;
 	ret = __recow_root(trans, info->chunk_root);
@@ -247,14 +299,9 @@ static int recow_roots(struct btrfs_trans_handle *trans,
 	ret = __recow_root(trans, info->dev_root);
 	if (ret)
 		return ret;
-	ret = __recow_root(trans, info->csum_root);
+	ret = recow_global_roots(trans);
 	if (ret)
 		return ret;
-	if (info->free_space_root) {
-		ret = __recow_root(trans, info->free_space_root);
-		if (ret)
-			return ret;
-	}
 	return 0;
 }
 
@@ -301,10 +348,10 @@ static int create_one_raid_group(struct btrfs_trans_handle *trans,
 
 static int create_raid_groups(struct btrfs_trans_handle *trans,
 			      struct btrfs_root *root, u64 data_profile,
-			      u64 metadata_profile, int mixed,
+			      u64 metadata_profile, bool mixed,
 			      struct mkfs_allocation *allocation)
 {
-	int ret;
+	int ret = 0;
 
 	if (metadata_profile) {
 		u64 meta_flags = BTRFS_BLOCK_GROUP_METADATA;
@@ -331,7 +378,6 @@ static int create_raid_groups(struct btrfs_trans_handle *trans,
 		if (ret)
 			return ret;
 	}
-	ret = recow_roots(trans, root);
 
 	return ret;
 }
@@ -421,14 +467,14 @@ static int zero_output_file(int out_fd, u64 size)
 	return ret;
 }
 
-static int is_ssd(const char *file)
+static bool is_ssd(const char *file)
 {
 	char rotational;
 	int ret;
 
 	ret = device_get_queue_param(file, "rotational", &rotational, 1);
 	if (ret < 1)
-		return 0;
+		return false;
 
 	return rotational == '0';
 }
@@ -550,7 +596,7 @@ static int cleanup_temp_chunks(struct btrfs_fs_info *fs_info,
 {
 	struct btrfs_trans_handle *trans = NULL;
 	struct btrfs_block_group_item *bgi;
-	struct btrfs_root *root = fs_info->extent_root;
+	struct btrfs_root *root = btrfs_extent_root(fs_info, 0);
 	struct btrfs_key key;
 	struct btrfs_key found_key;
 	struct btrfs_path path;
@@ -879,19 +925,20 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 	u32 nodesize = 0;
 	u32 sectorsize = 0;
 	u32 stripesize = 4096;
-	int zero_end = 1;
+	bool zero_end = true;
 	int fd = -1;
 	int ret = 0;
 	int close_ret;
 	int i;
-	int mixed = 0;
-	int nodesize_forced = 0;
-	int data_profile_opt = 0;
-	int metadata_profile_opt = 0;
-	int discard = 1;
-	int ssd = 0;
-	int zoned = 0;
-	int force_overwrite = 0;
+	bool mixed = false;
+	bool nodesize_forced = false;
+	bool data_profile_opt = false;
+	bool metadata_profile_opt = false;
+	bool discard = true;
+	bool ssd = false;
+	bool zoned = false;
+	bool force_overwrite = false;
+	int oflags;
 	char *source_dir = NULL;
 	bool source_dir_set = false;
 	bool shrink_rootdir = false;
@@ -902,7 +949,7 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 	int saved_optind;
 	char fs_uuid[BTRFS_UUID_UNPARSED_SIZE] = { 0 };
 	u64 features = BTRFS_MKFS_DEFAULT_FEATURES;
-	u64 runtime_features = 0;
+	u64 runtime_features = BTRFS_MKFS_DEFAULT_RUNTIME_FEATURES;
 	struct mkfs_allocation allocation = { 0 };
 	struct btrfs_mkfs_config mkfs_cfg;
 	enum btrfs_csum_type csum_type = BTRFS_CSUM_TYPE_CRC32;
@@ -947,28 +994,28 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 			break;
 		switch(c) {
 			case 'f':
-				force_overwrite = 1;
+				force_overwrite = true;
 				break;
 			case 'd':
 				data_profile = parse_profile(optarg);
-				data_profile_opt = 1;
+				data_profile_opt = true;
 				break;
 			case 'l':
 				warning("--leafsize is deprecated, use --nodesize");
 				/* fall through */
 			case 'n':
 				nodesize = parse_size_from_string(optarg);
-				nodesize_forced = 1;
+				nodesize_forced = true;
 				break;
 			case 'L':
 				label = parse_label(optarg);
 				break;
 			case 'm':
 				metadata_profile = parse_profile(optarg);
-				metadata_profile_opt = 1;
+				metadata_profile_opt = true;
 				break;
 			case 'M':
-				mixed = 1;
+				mixed = true;
 				break;
 			case 'O': {
 				char *orig = strdup(optarg);
@@ -1012,7 +1059,7 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 				break;
 			case 'b':
 				block_count = parse_size_from_string(optarg);
-				zero_end = 0;
+				zero_end = false;
 				break;
 			case 'v':
 				bconf_be_verbose();
@@ -1030,7 +1077,7 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 					BTRFS_UUID_UNPARSED_SIZE - 1);
 				break;
 			case 'K':
-				discard = 0;
+				discard = false;
 				break;
 			case 'q':
 				bconf_be_quiet();
@@ -1066,7 +1113,7 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 	if (dev_cnt == 0)
 		print_usage(1);
 
-	zoned = (features & BTRFS_FEATURE_INCOMPAT_ZONED);
+	zoned = !!(features & BTRFS_FEATURE_INCOMPAT_ZONED);
 
 	if (source_dir_set && dev_cnt > 1) {
 		error("the option -r is limited to a single device");
@@ -1118,7 +1165,7 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 			printf(
 	"Zoned: %s: host-managed device detected, setting zoned feature\n",
 			       file);
-		zoned = 1;
+		zoned = true;
 		features |= BTRFS_FEATURE_INCOMPAT_ZONED;
 	}
 
@@ -1130,19 +1177,10 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 		u64 tmp;
 
 		if (!metadata_profile_opt) {
-			if (dev_cnt == 1 && ssd && bconf.verbose)
-				printf("Detected a SSD, turning off metadata "
-				"duplication.  Mkfs with -m dup if you want to "
-				"force metadata duplication.\n");
-
-			if (dev_cnt > 1) {
+			if (dev_cnt > 1)
 				tmp = BTRFS_MKFS_DEFAULT_META_MULTI_DEVICE;
-			} else {
-				if (ssd)
-					tmp = BTRFS_MKFS_DEFAULT_META_ONE_DEVICE_SSD;
-				else
-					tmp = BTRFS_MKFS_DEFAULT_META_ONE_DEVICE;
-			}
+			else
+				tmp = BTRFS_MKFS_DEFAULT_META_ONE_DEVICE;
 			metadata_profile = tmp;
 		}
 		if (!data_profile_opt) {
@@ -1185,6 +1223,12 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 		features |= BTRFS_FEATURE_INCOMPAT_RAID1C34;
 	}
 
+	/* Extent tree v2 comes with a set of mandatory features. */
+	if (features & BTRFS_FEATURE_INCOMPAT_EXTENT_TREE_V2) {
+		features |= BTRFS_FEATURE_INCOMPAT_NO_HOLES;
+		runtime_features |= BTRFS_RUNTIME_FEATURE_FREE_SPACE_TREE;
+	}
+
 	if (zoned) {
 		if (source_dir_set) {
 			error("the option -r and zoned mode are incompatible");
@@ -1305,20 +1349,24 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 	if (ret)
 		goto error;
 
-	if (zoned && ((metadata_profile | data_profile) &
-		      BTRFS_BLOCK_GROUP_PROFILE_MASK)) {
+	if (zoned && (!zoned_profile_supported(metadata_profile) ||
+		      !zoned_profile_supported(data_profile))) {
 		error("zoned mode does not yet support RAID/DUP profiles, please specify '-d single -m single' manually");
 		goto error;
 	}
 
 	dev_cnt--;
 
+	oflags = O_RDWR;
+	if (zoned && zoned_model(file) == ZONED_HOST_MANAGED)
+		oflags |= O_DIRECT;
+
 	/*
 	 * Open without O_EXCL so that the problem should not occur by the
 	 * following operation in kernel:
 	 * (btrfs_register_one_device() fails if O_EXCL is on)
 	 */
-	fd = open(file, O_RDWR);
+	fd = open(file, oflags);
 	if (fd < 0) {
 		error("unable to open %s: %m", file);
 		goto error;
@@ -1349,6 +1397,13 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 	    btrfs_bg_type_to_tolerated_failures(data_profile))
 		warning("metadata has lower redundancy than data!\n");
 
+	printf("NOTE: several default settings have changed in version 5.15, please make sure\n");
+	printf("      this does not affect your deployments:\n");
+	printf("      - DUP for metadata (-m dup)\n");
+	printf("      - enabled no-holes (-O no-holes)\n");
+	printf("      - enabled free-space-tree (-R free-space-tree)\n");
+	printf("\n");
+
 	mkfs_cfg.label = label;
 	memcpy(mkfs_cfg.fs_uuid, fs_uuid, sizeof(mkfs_cfg.fs_uuid));
 	mkfs_cfg.num_bytes = dev_block_count;
@@ -1358,7 +1413,10 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 	mkfs_cfg.features = features;
 	mkfs_cfg.runtime_features = runtime_features;
 	mkfs_cfg.csum_type = csum_type;
-	mkfs_cfg.zone_size = zone_size(file);
+	if (zoned)
+		mkfs_cfg.zone_size = zone_size(file);
+	else
+		mkfs_cfg.zone_size = 0;
 
 	ret = make_btrfs(fd, &mkfs_cfg);
 	if (ret) {
@@ -1472,6 +1530,33 @@ raid_groups:
 		goto out;
 	}
 
+	/*
+	 * Commit current transaction so we can COW all existing tree blocks
+	 * to newly created raid groups.
+	 * As currently we use btrfs_search_slot() to COW tree blocks in
+	 * recow_roots(), if a tree block is already modified in current trans,
+	 * it won't be re-COWed, thus it will stay in temporary chunks.
+	 */
+	ret = btrfs_commit_transaction(trans, root);
+	if (ret) {
+		errno = -ret;
+		error("unable to commit transaction before recowing trees: %m");
+		goto out;
+	}
+	trans = btrfs_start_transaction(root, 1);
+	if (IS_ERR(trans)) {
+		errno = -PTR_ERR(trans);
+		error("failed to start transaction: %m");
+		goto error;
+	}
+	/* COW all tree blocks to newly created chunks */
+	ret = recow_roots(trans, root);
+	if (ret) {
+		errno = -ret;
+		error("unable to COW tree blocks to new profiles: %m");
+		goto out;
+	}
+
 	ret = create_data_reloc_tree(trans);
 	if (ret) {
 		error("unable to create data reloc tree: %d", ret);
